
#include "webkit_server.h"

__attribute__((visibility("hidden")))extern trace g_t1;
using namespace std;

webkit_server::webkit_server(PF* arr, std::string sock): func_array(arr), socket_path(sock){

    g_t1.logInfo("Inside webkit_server contructor\n");;
    pthread_attr_t tattr;
    pthread_attr_init(&tattr);
    pthread_create(&recv_thread_id, &tattr, &webkit_server::initCommWrapper, (void*)this);
    pthread_attr_destroy(&tattr);
}

void webkit_server::deinitComm(){

    void* res;
    g_t1.logInfo("WEBKIT:%d webkit_server::deinitComm Entered\n", getpid());

    //Also send a singal to recvMessage thread that the communication is closed
    if(pthread_cancel(recv_thread_id)){
        g_t1.logInfo("WEBKIT:%d Error in cancelling the thread in deinitComm!", getpid());
        g_t1.logInfo("%s\n", strerror(errno));
    }
    
    pthread_join(recv_thread_id, &res);
    if(res != PTHREAD_CANCELED){
        g_t1.logInfo("WEBKIT:%d recv_thread_id did not cancel succesfully! Not expected behaviour");
    }
    if(client_fd != -1){
        int err = close(client_fd);
        if(err == -1){
            g_t1.logInfo("WEBKIT:%d webkit_server::deinitComm could not close the client_fd, error: %s\n", getpid(), strerror(errno));
        }
        client_fd = -1;
    }
    else{
        g_t1.logInfo("WEBKIT:%d webkit_server::deinitComm client_fd is already closed", getpid());
    }
    
    if(client_pid > 0 ){

        //Send a kill singal to the extension process
        kill(client_pid, SIGTERM);

        //Block until the extension process is killed
        while(  (kill(client_pid, 0) == 0)  ){
            g_t1.logInfo("WEBKIT:%d Waiting for the extension process %d to be killed\n", getpid(), client_pid);
            usleep(2000); //Wait for 2 milliseconds 
        }

        assert(errno == ESRCH);
    }
    else{
        g_t1.logInfo("WEBKIT:%d webkit_server::client_pid is 0 or -1, could not connect to extension\n", getpid());
    }


}

/*static */ void* webkit_server::initCommWrapper(void* ws)
{
    webkit_server* t_ws = (webkit_server*)ws;
    g_t1.logInfo("WEBKIT:%d Entered webkit_server::initCommWrapper()\n", getpid());
    t_ws->sinitComm();
    return NULL;
}



//Converts serailized args to int[] and char[] and sends it over
void webkit_server::processMessage(webkit_message *&s)
{

    int args1[s->total_int_args];
    std::string args2[s->total_string_args];

    int cur = 0;
    char *buf = (char *)malloc(BUFSIZE);

    for (int i = 0; i < s->total_int_args; i++)
    {
        args1[i] = webkit_util::parseNextInt(cur, s->serialized_arg, buf);
    }

    for (int i = 0; i < s->total_string_args; i++)
    {
        args2[i] = webkit_util::parseNextString(cur, s->serialized_arg, buf);
    }


    //If pid of ext is sent, handle it in webkit_server class itself
    if(s->func_name == enum_pidOfExt){
        client_pid = args1[0]; //Extension pid 
    }
    else{
        func_array[s->func_name](args1, args2);
    }

    free(buf);
}

/*
Note: This function is now thread safe. As it is a different object
*/
void webkit_server::recvMessage()
{
    int total_len_of_string;

    g_t1.logInfo("WEBKIT:%d webkit_server::recvMessage this is the thread id: %d\n" , getpid() , pthread_self() );
    webkit_message *s; 
    int wrv;
    char *buf1 = (char *)malloc(BUFSIZE);
    if(buf1 == NULL){
        g_t1.logInfo("WEBKIT:%d webkit_server::recvMessage, Could not malloc, exiting error: %s\n", getpid(), strerror(errno));
        perror("Couldn't malloc, failing\n");
        exit(-1);
    }

    if (client_fd == -1)
    {
        g_t1.logInfo("WEBKIT:%d webkit_server::recvMessage, client_fd is -1, not connected to extension process yet", getpid());
        perror("webkit_server::recvMessage Error in client_fd\n");
        return;
    }

    memset(buf1, 0, BUFSIZE);

    while (1)
    {

        while (1)
        {
            if(client_fd == -1){
                g_t1.logInfo("WEBKIT:%d webkit_server::recvMessage, client_fd is -1, exiting thread\n", getpid());
                pthread_exit(NULL);
            }
            wrv = read(client_fd, buf1, 4 * sizeof(int));
            if (wrv == 0)
            {
                continue;
            }
            else
            {
                break;
            }
        }
        if (wrv != (4 * sizeof(int)))
        {
            g_t1.logInfo("WEBKIT:%d webkit_server::recvMessage error in reading, exiting\n", getpid());
            exit(-1);
        }

        //Read first four ints succesfully
        total_len_of_string = *(int *)(buf1 + 3 * sizeof(int));

        assert(total_len_of_string < BUFSIZE);


        s = (webkit_message *)malloc(sizeof(webkit_message) + total_len_of_string);

        s->func_name = *(func_enum *)buf1;
        s->total_int_args = *((int *)((char *)buf1 + 1 * sizeof(int)));
        s->total_string_args = *((int *)((char *)buf1 + 2 * sizeof(int)));
        s->total_len_of_string = *(int *)((char *)buf1 + 3 * sizeof(int));

        int to_be_read = s->total_len_of_string;
        
        if(s->total_len_of_string != 0){

            //Read until everything from the socket is read
            while(to_be_read != 0){
            
                wrv = read(client_fd, (s->serialized_arg) + (s->total_len_of_string - to_be_read), to_be_read);
                if (wrv == 0)
                {
                    continue;
                }
                else
                {
                    if(wrv < 0){
                        g_t1.logInfo("WEBKIT:%d webkit_client::recvMessage Error in reading in webkit_client %s \n", getpid(), strerror(errno));
                        exit(-1);
                    }
                    else{
                        //partially read, reading more\n";
                        //Place the first char from the position where it was filled last time
                        to_be_read = to_be_read - wrv;
                        assert(to_be_read >=0);
                    }
                }
            }
        }
        processMessage(s);
        free(s); 

    }

    free(buf1); 

}

void webkit_server::sendMessage(webkit_message *&s)
{

    if (client_fd == -1)
    {
        perror("client_fd not initialized yet, exiting\n");
        g_t1.logInfo("WEBKIT:%d webkit_server::sendMessage, client_fd not initialized yet, exiting\n", getpid());
        exit(-1);
    }

    char *buf = (char *)malloc(BUFSIZE);
    int wrv;
    int size_struct = sizeof(webkit_message) + s->total_len_of_string;

    if (size_struct > BUFSIZE)
    {
        g_t1.logInfo("WEBKIT:%d webkit_server::sendMessage, size_struct > BUFSIZE error, exiting\n", getpid());
        exit(-1);
    }
    memcpy(buf, s, size_struct);

    wrv = write(client_fd, buf, size_struct);
    if (wrv != size_struct)
    {
        if (wrv > 0)
        {
            g_t1.logInfo("WEBKIT:%d webkit_server::sendMessage, could not write whole message, exiting\n"), getpid();
            exit(-1);
        }
        else
        {
            perror("Error in writing the message\n");
            g_t1.logInfo("WEBKIT:%d webkit_server::sendMessage, could not write the message error: %s, exiting\n", strerror(errno), getpid());

            exit(-1);
        }
    }

    free(buf);
}

/*
 This function waits until a client(webextension) connects to it and then moves forward to recvMessage.
 recvMessage run in an infinte loop to receive any incomming messages
*/
void* webkit_server::sinitComm()
{

    struct sockaddr_un addr;
    const char* sock_char = socket_path.c_str();
    int fd;

    //Create the socket named fd
    if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1)
    {
        g_t1.logInfo("WEBKIT:%d webkit_server::sinitComm, Could not create a socket: %s, exiting\n", strerror(errno), getpid());
        perror("socket error");
        exit(-1);
    }

    memset(&addr, 0, sizeof(addr));
    addr.sun_family = AF_UNIX;
    strncpy(addr.sun_path, sock_char, sizeof(addr.sun_path) - 1);
    unlink(sock_char); //Deletes the path if it exists

    //Bind the fd with the webkit_socket path
    if (bind(fd, (struct sockaddr *)&addr, sizeof(addr)) == -1)
    {
        g_t1.logInfo("WEBKIT:%d webkit_server::sinitComm, Could not bind to the socket: %s, exiting\n", getpid(), strerror(errno));
        perror("bind error\n");
        exit(-1);
    }

    //Listen until someone tries to connect
    if (listen(fd, 1) == -1)
    {
        g_t1.logInfo("WEBKIT:%d webkit_server::sinitComm, listen error: %s, exiting\n", getpid(), strerror(errno));
        perror("listen error");
        exit(-1);
    }


    //get the fd of the client
    client_fd = accept(fd, NULL, NULL);
    if(client_fd == -1)
    {
        g_t1.logInfo("WEBKIT:%d webkit_server::sinitComm, accept error: %s, exiting thread\n", getpid(), strerror(errno));
        perror("accept error");
        pthread_exit(NULL);
    }

    //An infinite loop until pthread_cancel is called
    recvMessage();

    return NULL;
}